@@ -1,6 +1,6 @@
 package Apache::GD::Graph;
 
-$VERSION = 0.92;
+$VERSION = 0.93;
 
 =head1 NAME
 
@@ -13,31 +13,33 @@ In httpd.conf:
 	<Location /chart>
 	SetHandler	perl-script
 	PerlHandler	+Apache::GD::Graph
-	# These are optional.
-	#PerlSetVar	Expires		30	# days default.
-	#PerlSetVar	CacheSize	5242880	# 5 megs default.
-	#PerlSetVar	ImageType	png	# Default.
-	# The default image type that graphs should be.
-	# png is default, gif requires <= GD 1.19.
-	# Any type supported by the installed version of GD will work.
+	## These are optional (defaults shown)
+	## 				In days:
+	#PerlSetVar	Expires		30
 	#
-	# The next specifies the quality of generated JPEG images, best not to
-	# specify this one and let GD figure it out.
+	##				In megs:
+	#PerlSetVar	CacheSize	5242880
+	#PerlSetVar	ImageType	png
 	#PerlSetVar	JpegQuality	75 # 0 to 100
-	# TTFFontPath tells GD::Text where to look for true type fonts. The
-	# example here is the default value used.
-	#PerlSetVar	TTFFontPath	/usr/ttfonts:/var/ttfonts:/usr/X11R6/lib/X11/fonts/ttf/:/usr/X11R6/lib/X11/fonts/truetype/:/usr/share/fonts/truetype
+	#PerlSetVar	TTFFontPath	/usr/ttfonts:
+	#/var/ttfonts:/usr/X11R6/lib/X11/fonts/ttf/:
+	#/usr/X11R6/lib/X11/fonts/truetype/:
+	#/usr/share/fonts/truetype
 	</Location>
 
 Then send requests to:
 
-	http://www.server.com/chart?type=lines&x_labels=[1st,2nd,3rd,4th,5th]&
-	data1=[1,2,3,4,5]&data2=[6,7,8,9,10]&dclrs=[blue,yellow,green]>
+	http://www.server.com/chart?type=lines&
+	x_labels=[1st,2nd,3rd,4th,5th]&
+	data1=[1,2,3,4,5]&
+	data2=[6,7,8,9,10]&
+	dclrs=[blue,yellow,green]>
 
 Options can also be sent as x-www-form-urlencoded data (ie., a form). This
-works better for large data sets, and allows simple charting forms to be set
-up. IE, for example, does not allow query strings larger than a kilobyte.
-Parameters in the query string take precedence over a form if specified.
+allows simple charting forms to be set up, also, Internet Explorer does not
+allow query strings larger than a kilobyte so in those cases a POSTED form must
+be used.  Parameters in the query string take precedence over a form if
+specified.
 
 =head1 INSTALLATION
 
@@ -56,9 +58,10 @@ with the chart creation process abstracted and placed on any server.
 For example, embedding a pie chart can be as simple as:
 
 	<img src="http://www.some-server.com/chart?type=pie&
-	x_labels=[greed,pride,wrath]&data1=[10,50,20]&dclrs=[green,purple,red]"
+	x_labels=[greed,pride,wrath]&data1=[10,50,20]&
+	dclrs=[green,purple,red]"
 	alt="pie chart of a few deadly sins">
-	<!-- Note that all of the above options are optional except for data1!  -->
+	<!-- All above options are optional except for data1 -->
 
 And it gets cached both server side, and along any proxies to the client, and
 on the client's browser cache. Not to mention, chart generation is
@@ -90,10 +93,27 @@ otherwise you will spend hours being very confused.
 
 GD::Graph has some options that take a font description, such as title_font,
 legend_font, etc. (these map to the appropriate set_FOO methods in GD::Graph,
-see that manpage). See the TTFFontPath variable under SYNOPSIS for how to set
-the search path for fonts. MAKE SURE your fonts are readable by the user the
-Apache server runs under, this is usually "www-data" or "nobody". Otherwise
-your fonts will mysteriously fail with no notice.
+see that manpage).
+
+The following fonts are built-in to GD, these strings will resolve into the
+appropriate fonts except when quoted:
+
+gdSmallFont, gdLargeFont, gdMediumBoldFont, gdTinyFont, gdGiantFont
+
+There is also a way to use your own True Type Fonts. See the TTFFontPath
+variable under SYNOPSIS for how to set the search path for fonts. MAKE SURE
+your fonts are readable by the user the Apache server runs under, this is
+usually "www-data" or "nobody". Otherwise your fonts will mysteriously fail
+with no notice.
+
+Fonts can also be specified as a relative path to the DocumentRoot of the
+server, these must begin with "../". For example, if you have a directory
+"fonts" under DocumentRoot, then you might specify a font like so:
+
+	../fonts/arial.ttf
+
+If DocumentRoot happens to be C</var/www> then the font that will be looked up
+is C</var/www/fonts/arial.ttf>.
 
 Sizes can be specified by using a list with the name and size. For example, if
 arial.ttf can be found somewhere in your TTFFontPath, you can do:
@@ -103,10 +123,22 @@ arial.ttf can be found somewhere in your TTFFontPath, you can do:
 To get a title using font Arial, in 20 points.
 
 Note that GD::Text does not parse out the names of fonts and such, you have to
-give it an actual filename, in the right case. So if using the Microsoft
-Windows core fonts, Arial Bold would be C<arialbd.ttf>. Here's an example:
+give it an actual filename, matches are case-insensitive. So if using the
+Microsoft Windows core fonts, Arial Bold would be C<arialbd.ttf>. Here's an
+example:
 
-	http://server/chart?data1=[1,2,3,4,5]&title_font=(arialbd.ttf,20)&title=Just%20A%20Line
+	http://server/chart?data1=[1,2,3,4,5]&
+	title_font=(arialbd.ttf,20)&
+	title=Just%20A%20Line
+
+=head1 COLORS
+
+All colors, including those specified for the captionN option, are specified
+using the colour names from L<GD::Graph::colour>. They are, at time of writing:
+
+white, lgray, gray, dgray, black, lblue, blue, dblue, gold, lyellow, yellow,
+dyellow, lgreen, green, dgreen, lred, red, dred, lpurple, purple, dpurple,
+lorange, orange, pink, dpink, marine, cyan, lbrown, dbrown.
 
 =head1 IMAGES
 
@@ -185,22 +217,50 @@ can be of any type your copy of GD supports.
 =item B<captionN>
 
 Draws a character string using a TrueType font at an arbitrary location.  Takes
-an array of C<($fgcolor,$fontname,$ptsize,$angle,$x,$y,$string)> where $fgcolor
-is the foreground color, $fontname is the name of a TTF font see L</FONTS> ,
-$ptsize is the point size, $x and $y are the coordinates, and $string is the
-actual characters to draw.
+an array of
+C<($fgcolor,$fontname,$ptsize,$angle,$x,$y,$string[,$box_color,$box_offset])>
+where $fgcolor is the foreground color, $fontname is the name of a TTF font see
+L</FONTS> , $ptsize is the point size, $x and $y are the coordinates, and
+$string is the actual characters to draw.
+
+$box_color and $box_offset are optional parameters, if set the caption will be
+drawn with a box around it in that color and that distance from the caption
+string. The default offset of 9 should work well in most cases.
 
 N is an integer from 1 onward, like for the dataN option. This lets you specify
 multiple strings to draw.
 
-This follows exactly the GD stringTTF method, see L<GD>.
+B<Note:> you cannot use builtin GD fonts like gdTinyFont for captions, you have
+to use a real TTF font.
+
+This uses the GD stringTTF method, see L<GD>. Colour names are indexed using
+the GD::Graph::colour builtins (see above), fonts are resolved by font path or
+relative to DocumentRoot, parameters are processed as per L</DATA TYPES>.
 
 Angle is in degrees, you will primarily use angle C<0> for normal left-to-right
 text. $x and $y are pixel coordinates from the upper left corner. $fontname is
-the name of a true-type font that will be found in the font path (see FONTS).
+the name of a true-type font that will be found in the font path L</FONTS>.
 Example:
 
-	http://isis/chart?data1=[1,2,3,4,5]&caption1=(1,arial.ttf,9,0,30,30,Hello)
+	http://isis/chart?data1=[1,2,3,4,5]&
+	caption1=(1,arial.ttf,9,0,30,30,Hello,red)
+
+To draw the box around the caption as a dashed or dotted line use:
+
+=item B<gd_set_style>
+
+This option sets the style for the special gdStyled color index. It's simply a
+list of colors that becomes the pattern for lines and such drawn with it. For
+example, to get a dashed red line:
+
+	gd_set_style=(red,red,red,red,red,red,
+	gdTransparent,gdTransparent,
+	gdTransparent,gdTransparent);
+
+The list can be arbitrarily long.
+
+B<Note:> at this time, the only place where you can use colors of this style is
+for the box around a caption. Just specify C<gdStyled> as the color.
 
 =item B<cache>
 
@@ -240,10 +300,12 @@ times with N increasing.
 =back
 
 ALL OTHER OPTIONS are passed to the corresponding set_<option> method, or the
-set(<option hash>) method using the following rules for the values:
+set(<option hash>) method using the following rules for the values:jj
 
 =over 8
 
+=head1 DATA TYPES
+
 =item B<undef>
 
 Becomes a real undef.
@@ -267,6 +329,17 @@ Becomes a hash reference.
 Is pulled into a file and the file name is passed to the respective option.
 (Can be any scheme besides http:// that LWP::Simple supports.)
 
+=item B<../fonts/arial.ttf>
+
+Paths following this pattern will be interpreted as paths relative to
+DocumentRoot of the web server.
+
+=item B<gdSmallFont, gdLargeFont, gdMediumBoldFont, gdTinyFont, gdGiantFont,
+	gdStyled, gdBrushed, gdStyledBrushed, gdTransparent>
+
+These are reserved strings. If not quoted, they will be converted to the
+builtin GD constants of the same name. See L<GD> for details.
+
 =item B<[undef,something,undef] or {key,undef}>
 
 You can create an array or hash with undefs.
@@ -279,27 +352,37 @@ inside arrays and hashes.
 DON'T USE SPACES, this is a common mistake. A space in a URL-encoded string is
 %20, or a + in a form.
 
-Nested structures are still not supported, maybe later.
-
 =back
 
 =cut
 
 use strict;
+use Data::Dumper;
 use Apache;
 use Apache::Constants qw/OK/;
 use HTTP::Date;
 use GD;
 use GD::Text;
 use GD::Graph;
-use GD::Graph::colour;
+use GD::Graph::colour qw(:convert);
 use File::Cache;
 
+use constant TRUE	=> 1;
+use constant FALSE	=> 0;
+
+use constant SECONDS_IN_DAY => 24 * 60 * 60;
+
 use constant EXPIRES	=> 30;
 use constant CACHE_SIZE	=> 5242880;
 use constant IMAGE_TYPE => 'png';
 use constant TTF_FONT_PATH	=> '/usr/ttfonts:/var/ttfonts:/usr/X11R6/lib/X11/fonts/ttf/:/usr/X11R6/lib/X11/fonts/truetype/:/usr/share/fonts/truetype';
 
+use constant DEFAULT_TYPE	=> 'lines';
+use constant DEFAULT_WIDTH	=> 400;
+use constant DEFAULT_HEIGHT	=> 300;
+
+use constant DEFAULT_CAPTION_BOX_OFFSET => 9;
+
 use constant TYPE_UNDEF		=> 0;
 use constant TYPE_SCALAR	=> 1;
 use constant TYPE_ARRAY		=> 2;
@@ -307,30 +390,71 @@ use constant TYPE_HASH		=> 3;
 use constant TYPE_URL		=> 4;
 use constant TYPE_LIST		=> 5;
 
-use constant STRIP_QUOTES => qr/['"]?(.*)['"]?/;
+use constant STRIP_QUOTES => qr/(?:['"]|\%22)?	# First quote char, optional
+				(.*)		# The main text.
+				(?:['"]|\%22)	# Second quote char
+				/x;
 
 use constant ARRAY_OPTIONS => qw(
 	dclrs borderclrs line_types markers types
 );
 
+use constant GD_CONSTANTS => {
+	gdSmallFont	=> gdSmallFont,
+	gdLargeFont	=> gdLargeFont,
+	gdMediumBoldFont=> gdMediumBoldFont,
+	gdTinyFont	=> gdTinyFont,
+	gdGiantFont	=> gdGiantFont,
+	gdStyled	=> gdStyled,
+	gdBrushed	=> gdBrushed,
+	gdStyledBrushed	=> gdStyledBrushed,
+	gdTransparent	=> gdTransparent
+};
+
 # Sub prototypes:
 
+sub init();
 sub handler ($);
 sub parse ($;$);
 sub arrayCheck ($$);
 sub error ($);
 sub makeDir ($);
-sub parseURL ($;$);
+sub parseElement ($;$);
 sub findFont ($);
+sub resolveColor ($$);
+
+# Package variables.
+
+my $first_request = TRUE;
+my ($r, $cache_size, $image_cache, $document_root, @cleanup_files);
 
 # Subs:
 
+# init()
+#
+# Called only once on the first request received. May be called once per child
+# in Apache.
+sub init() {
+# Set the GD::Text fontpath.
+	GD::Text->font_path ($r->dir_config('TTFFontPath') || TTF_FONT_PATH);
+
+	$cache_size = $r->dir_config('CacheSize') || CACHE_SIZE;
+
+	$image_cache = new File::Cache ({
+		namespace	=> 'Images',
+		max_size	=> $cache_size,
+		filemode	=> 0660
+	});
+
+	$document_root = $r->document_root;
+}
+
 sub handler ($) {
-	my $r = shift;
+	$r = shift;
 	$r->request($r);
 
-# Files to delete after request is processed.
-	my @cleanup_files;
+	init, $first_request = FALSE
+		if $first_request;
 
 	eval {
 		my $args = scalar $r->args;
@@ -351,27 +475,20 @@ EOF
 # Calculate Expires header based on either an "expires" parameter, the Expires
 # configuration variable (via PerlSetVar) or the EXPIRES constant, in days.
 # Then convert into seconds and round to an integer.
-		my $expires = exists $args{expires} ? $args{expires} :
-			      $r->dir_config('Expires') || EXPIRES;
-
-		$expires   *= 24 * 60 * 60;
-		$expires    = sprintf ("%d", $expires);
+		my $expires = exists $args{expires} ?
+			sprintf "%.0f", $args{expires} * SECONDS_IN_DAY
+			:
+			$r->dir_config('Expires') || EXPIRES;
 
 # Determine the type of image that the graph should be.
 # Allow an Accept: header with one specific image type to set it, a
 # PerlSetVar, or the image_type parameter.
 		my $image_type = lc($r->dir_config('ImageType')) || IMAGE_TYPE;
 
-# Set the GD::Text fontpath.
-		GD::Text->font_path(
-			$r->dir_config('TTFFontPath') ||
-			TTF_FONT_PATH
-		);
-
 		my $accepts_header = $r->header_in('Accept');
 		if (defined $accepts_header and
 		    $accepts_header =~ m!^\s*image/(\w+)\s*$!) {
-			my $image_type = $1;
+			$image_type = $1;
 		}
 
 		$image_type = $args{image_type} if $args{image_type};
@@ -380,7 +497,7 @@ EOF
 
 		die <<EOF unless GD::Image->can($image_type);
 The version of GD installed on this server does not support
-ImageType $image_type.
+image_type $image_type.
 EOF
 
 		my $jpeg_quality;
@@ -389,46 +506,31 @@ EOF
 					$r->dir_config('JpegQuality');
 		}
 
-		my $image_cache;
+		$args{cache} = TRUE if not exists $args{cache};
 
-		if (not exists $args{cache} or $args{cache} != 0) {
-			my $cache_size = $r->dir_config('CacheSize');
+		if ($args{cache} != FALSE && $cache_size > 0 &&
+		   defined(my $cached_image = $image_cache->get($args))) {
+			$r->header_out (
+				"Expires" => time2str(time + $expires)
+			);
+			$r->send_http_header (
+				"image/$image_type"
+			);
+			$r->print($cached_image);
 
-			if (defined $cache_size and $cache_size != 0) {
-				$image_cache = new File::Cache ( {
-					namespace	=> 'Images',
-					max_size	=> $cache_size ||
-								CACHE_SIZE,
-					filemode	=> 0660
-				} );
-
-				if (my $cached_image =
-				    $image_cache->get($args)) {
-					$r->header_out (
-						"Expires" => time2str (
-							time + $expires
-						)
-					);
-					$r->send_http_header (
-						"image/$image_type"
-					);
-					$r->print($cached_image);
-
-					return OK;
-				}
-			}
+			return OK;
 		}
 
-		my $type   = delete $args{type}   || 'lines';
-		my $width  = delete $args{width}  || 400;
-		my $height = delete $args{height} || 300;
+		my $type   = delete $args{type}   || DEFAULT_TYPE;
+		my $width  = delete $args{width}  || DEFAULT_WIDTH;
+		my $height = delete $args{height} || DEFAULT_HEIGHT;
 
 		$type =~ m/^(\w+)$/;
 		$type = $1;	# untaint it!
 
 		my @data;
-		my $ctr = 1;
-		my $key = "data$ctr";
+		my $i = 1;
+		my $key = "data$i";
 		while (exists $args{$key}) {
 			my ($type, $array, @rest) = (parse delete $args{$key});
 			if ($type == TYPE_LIST) {
@@ -436,7 +538,7 @@ EOF
 			}
 			arrayCheck $key, $array;
 			push @data, $array;
-			$key = "data".(++$ctr);
+			$key = "data".(++$i);
 		}
 
 		die "Please supply at least a data1 argument."
@@ -451,9 +553,9 @@ EOF
 			delete $args{no_axes};
 		}
 
-		my ($x_labels, $x_labels_type);
+		my $x_labels;
 		if (exists $args{x_labels}) {
-			($x_labels, $x_labels_type) =
+			$x_labels =
 				(parse delete $args{x_labels})[1];
 		} else {
 			$x_labels = undef;
@@ -495,7 +597,9 @@ Could not create an instance of class GD::Graph::$type: $@
 EOF
 		}
 
-		my $to_file = delete $args{to_file};
+		my $to_file = (parseElement delete $args{to_file})[1];
+# Untaint it!
+		($to_file) = ($to_file =~ /([\w.\/]+)/);
 
 		for my $option (keys %args) {
 			my ($type, $value, @rest) = parse ($args{$option});
@@ -511,11 +615,6 @@ EOF
 
 			arrayCheck $option, $value
 				if index (ARRAY_OPTIONS, $option) != -1;
-
-			if ($type == TYPE_URL) {
-				push @cleanup_files, $args{$option};
-			}
-
 		};
 
 # Check if background image specified.
@@ -536,40 +635,70 @@ EOF
 
 # Check if we need to draw captions, draw them after graph is plotted.
 		my @captions;
-		$ctr = 1;
-		$key = "caption$ctr";
+		$i = 1;
+		$key = "caption$i";
 		while (exists $args{$key}) {
 			die <<EOF unless UNIVERSAL::isa($args{$key}, 'ARRAY');
 Caption must be an array. See the Apache::GD::Graph man page or the StringTTF
 method in the GD man page for details.
 EOF
-			push @captions, $args{$key};
-			delete $args{$key};
-			$key = "caption".(++$ctr);
+			push @captions, delete $args{$key};
+			$key = "caption".(++$i);
 		}
 
+# Style for the special gdStyled color.
+		my $gd_style = delete $args{gd_set_style};
+
 		$graph->set(%args);
 
-		my $result = $graph->plot([$x_labels, @data]);
+		my $image = $graph->plot([$x_labels, @data])
+			or die <<EOF;
+Could not create graph: @{[ $graph->error ]}
+EOF
+
+		$image->setStyle (
+			map { resolveColor ($graph => $_) }
+			@$gd_style
+		) if $gd_style;
 
 # Draw captions.
 		for my $caption (@captions) {
 			undef $@;
+
+# Argument 1 is the color, have to resolve GD::Graph::colour builtins into
+# indexes on the GD image.
+			$caption->[0] = resolveColor($graph => $caption->[0]);
+
 # Argument 2 to caption is the font name, GD expects a full path.
 			$caption->[1] = findFont($caption->[1]);
-			my @bounds = $result->stringTTF(@$caption);
+
+			my @bounds = $image->stringTTF(@$caption[0..6]);
+
 			die "Could not draw caption: @{[ join ', ', @$caption ]}: $@" if $@;
+
+# Draw box around caption.
+			next unless defined(my $box_clr = $caption->[7]);
+
+			my $offset = defined $caption->[8] ?
+					$caption->[8] :
+					DEFAULT_CAPTION_BOX_OFFSET;
+# Upper left.
+			$bounds[6] -= $offset;
+			$bounds[7] -= $offset;
+# Lower right.
+			$bounds[2] += $offset;
+			$bounds[3] += $offset;
+
+			$image->rectangle(
+				@bounds[6,7,2,3],
+				resolveColor($graph => $box_clr)
+			);
 		}
 
-		die <<EOF if not defined $result;
-Could not create graph: @{[ $graph->error ]}
-EOF
-
-		my $image;
 		if (defined $jpeg_quality) {
-			$image = $result->jpeg($jpeg_quality);
+			$image = $image->jpeg($jpeg_quality);
 		} else {
-			$image = $result->$image_type();
+			$image = $image->$image_type();
 		}
 
 		if (not $to_file) {
@@ -577,10 +706,11 @@ EOF
 			$r->send_http_header("image/$image_type");
 			$r->print($image);
 
-			$image_cache->set($args, $image) if defined $image_cache;
+			$image_cache->set($args, $image) if $args{cache};
 		} else {
 			my $destination = new IO::File ">$to_file"
-				or die "Could not writ to $to_file: $!";
+				or die "Could not write to_file $to_file: $!";
+
 			print $destination $image;
 
 			$r->send_http_header("text/plain");
@@ -611,12 +741,12 @@ EOF
 # undef			-- a real undef
 # foo_bar		-- a scalar
 # [1,2,undef,"foo",bar]	-- an array
-# [3,4,undef,"baz"]	-- a list
+# (3,4,undef,"baz")	-- a list
 # {1,2,'3',foo}		-- a hash
-# or
 # http://some/url.png	-- pull a URL into a file, returning that. The file
 # will be relative to a directory given as the second parameter, or /tmp if not
 # specified.
+# ../some/file		-- a file relative to DocumentRoot
 sub parse ($;$) {
 	local $_ = shift;
 	my $dir  = shift || '/tmp';
@@ -624,35 +754,43 @@ sub parse ($;$) {
 	return (TYPE_UNDEF, undef) if $_ eq 'undef';
 
 	if (/^\[(.*)\]$/) {
-		return (TYPE_ARRAY, [ map { $_ eq 'undef' ? undef : (parseURL $_, $dir)[1] }
+		return (TYPE_ARRAY, [ map { $_ eq 'undef' ? undef : (parseElement $_, $dir)[1] }
 				split /,/, $1, -1
 		        ]);
 	}
 
 	if (/^\{(.*)\}$/) {
-		return (TYPE_HASH, { map { $_ eq 'undef' ? undef : (parseURL $_, $dir)[1] }
+		return (TYPE_HASH, { map { $_ eq 'undef' ? undef : (parseElement $_, $dir)[1] }
 				split /,/, $1, -1
 		        });
 	}
 
 	if (/^\((.*)\)$/) {
-		return (TYPE_LIST, map { $_ eq 'undef' ? undef : (parseURL $_, $dir)[1] }
+		return (TYPE_LIST, map { $_ eq 'undef' ? undef : (parseElement $_, $dir)[1] }
 				split /,/, $1, -1
 		       );
 	}
 
-	return parseURL $_, $dir;
+	return parseElement $_, $dir;
 }
 
-# parseURL ($value)
+# parseElement ($value)
 #
 # First strips quotes off the ends of $value.  Then checks whether $value is a
 # URL, and if so, fetches it into a file and returns the (TYPE_URL, file_name),
 # otherwise returns (TYPE_SCALAR, $value).
-sub parseURL ($;$) {
+#
+# Will also parse paths relative to DocumentRoot, for example
+# ../fonts/arial.ttf.
+sub parseElement ($;$) {
 	$_	= shift;
 	my $dir	= shift || '/tmp';
-	($_) = (/@{[STRIP_QUOTES]}/);
+
+	if (defined(my $constant = GD_CONSTANTS->{$_})) {
+		return (TYPE_SCALAR, $constant)
+	}
+
+	$_ = $1 if /@{[STRIP_QUOTES]}/;
 
 	if (m!^\w+://!) {
 		use LWP::Simple;
@@ -671,6 +809,13 @@ Could not retrieve data from: $url
 EOF
 
 		print $file $contents;
+
+		push @cleanup_files, $file;
+
+		return (TYPE_URL, $file_name);
+	} elsif (s!^\.\./!!) {
+		my $file_name = $document_root.'/'.$_;
+
 		return (TYPE_URL, $file_name);
 	} else {
 		return (TYPE_SCALAR, $_);
@@ -694,7 +839,6 @@ sub error ($) {
 	my $message	= shift;
 # Ending newlines look ugly in the error log.
 	chomp $message;
-	my $r		= Apache->request;
 	my $contact	= $r->server->server_admin;
 	$r->send_http_header("text/html");
 	$r->print(<<"EOF");
@@ -724,16 +868,34 @@ EOF
 #
 # Returns undef if no font was found.
 sub findFont ($) {
-	my $name = shift;
+	my $name = shift || return undef;
+
+# Don't need to search for qualified file names or font objects.
+	return $name
+		if ($name =~ m!^/!) || (ref $name);
+
 	my @path = map { m!(.*?)/*$! } split /:/, GD::Text->font_path;
 
 	for my $path (@path) {
 		for my $font (<$path/*>) {
-			return $font if $font =~ m!/$name$!;
+			return $font if $font =~ m!/$name$!i;
 		}
 	}
 }
 
+# resolveColor ($gd_graph_object, $color_name)
+#
+# Resolve a GD::Graph::colour builtin into the index for GD, if it's not
+# numeric already.
+#
+sub resolveColor ($$) {
+	my ($graph, $color) = @_;
+
+	return $color if $color !~ /[A-z]/;
+
+	return $graph->set_clr (GD::Graph::colour::_rgb($color));
+}
+
 1;
 
 __END__
@@ -768,15 +930,14 @@ option.
 
 =head1 TODO
 
-Boxed captions.
 Variable mapping of x-labels to data points.
-If possible, a comprehensive test suite.
-Make it faster?
+Better test suite.
 
 =head1 SEE ALSO
 
 L<perl>,
 L<GD::Graph>,
+L<GD::Graph::colour>,
 L<GD>
 
 =cut
